@@ -1,18 +1,71 @@
 -- This is part of the OpTeX project, see http://petr.olsak.net/optex
 
--- The basic lua functions and declarations used in \OpTeX/ are here
+-- Basic Lua functions and declarations used in \OpTeX/ are defined here
+--
+-- Local imports of standard Lua or \LuaTeX/ functions and constants are the first thing we do.
+-- They mostly serve one of several purposes:
+--
+-- \begitems
+-- * shorthands for commonly used functions,
+-- * speed of access for commonly executed functions,
+-- * common constants to be used consistently,
+-- * use of local definition that can't be externally overwritten.
+-- \enditems
 
 local fmt = string.format
 
+local node_id = node.id
+local node_subtype = node.subtype
+local glyph_id = node_id("glyph")
+local rule_id = node_id("rule")
+local glue_id = node_id("glue")
+local hlist_id = node_id("hlist")
+local vlist_id = node_id("vlist")
+local disc_id = node_id("disc")
+local kern_id = node_id("kern")
+local whatsit_id = node_id("whatsit")
+local pdfliteral_id = node_subtype("pdf_literal")
+local pdfsave_id = node_subtype("pdf_save")
+local pdfrestore_id = node_subtype("pdf_restore")
+local token_getmacro = token.get_macro
+
+local direct = node.direct
+local todirect = direct.todirect
+local tonode = direct.tonode
+local getfield = direct.getfield
+local setfield = direct.setfield
+local getwhd = direct.getwhd
+local getid = direct.getid
+local getlist = direct.getlist
+local setlist = direct.setlist
+local getleader = direct.getleader
+local getfont = direct.getfont
+local getattribute = direct.get_attribute
+local setattribute = direct.set_attribute
+local insertbefore = direct.insert_before
+local copy = direct.copy
+local traverse = direct.traverse
+local one_bp = tex.sp("1bp")
+
+local tex_print = tex.print
+local tex_setbox = tex.setbox
+local tex_getcount = tex.getcount
+local tex_setcount = tex.setcount
+local token_scanint = token.scan_int
+local token_scanlist = token.scan_list
+local token_setmacro = token.set_macro
+local font_getfont = font.getfont
+--
 -- \medskip\secc General^^M
 --
--- Define namespace where some \OpTeX/ functions will be added.
+-- Define namespace where \"public" \OpTeX/ functions will be added.
 
 local optex = _ENV.optex or {}
 _ENV.optex = optex
 
 -- Error function used by following functions for critical errors.
-local function err(message)
+local function err(...)
+    local message = fmt(...)
     error("\nerror: "..message.."\n")
 end
 --
@@ -30,9 +83,30 @@ function optex.mdfive(file)
     if fh then
         local data = fh:read("*a")
         fh:close()
-        tex.print(md5.sumhexa(data))
+        tex_print(md5.sumhexa(data))
    end
 end
+
+-- The `optex.raw_ht ()` function measures the height plus depth of given
+-- `\vbox` saved by `\setbox`. It solves the \"dimension too large" problem with big boxes.
+-- It is used in the \^`\_rawht` macro.
+
+function optex.raw_ht()
+    local nod = tex.box[token.scan_int()]       -- read head node of the given box
+    local ht = 0
+    assert(nod ~= nil and nod.id == vlist_id, "given box must be a \\vbox")
+    for n in node.traverse(nod.head) do
+        if n.id == hlist_id or n.id == vlist_id or n.id == rule_id then
+            ht = ht + n.height + n.depth
+        elseif n.id == glue_id then
+            ht = ht + n.width
+        elseif n.id == kern_id then
+            ht = ht + n.kern
+        end
+    end
+    tex_print(fmt('%.0f', ht/65536))
+end
+
 --
 -- \medskip\secc[lua-alloc] Allocators^^M
 local alloc = _ENV.alloc or {}
@@ -192,11 +266,11 @@ local valid_callback_types = {
 -- `call_callback`. A default function is only needed by "exclusive" callbacks.
 function callback.create_callback(name, cbtype, default)
     if callback_types[name] then
-        err("cannot create callback '"..name.."' - it already exists")
+        err("cannot create callback '%s' - it already exists", name)
     elseif not valid_callback_types[cbtype] then
-        err("cannot create callback '"..name.. "' with invalid callback type '"..cbtype.."'")
-    elseif ctype == "exclusive" and not default then
-        err("unable to create exclusive callback '"..name.."', default function is required")
+        err("cannot create callback '%s' with invalid callback type '%s'", name, cbtype)
+    elseif cbtype == "exclusive" and not default then
+        err("unable to create exclusive callback '%s', default function is required", name)
     end
 
     callback_types[name] = cbtype
@@ -227,7 +301,27 @@ function callback.add_to_callback(name, fn, description)
             return call_callback(name, ...)
         end)
     else
-        err("cannot add to callback '"..name.."' - no such callback exists")
+        err("cannot add '%s' to callback '%s' - no such callback exists", description, name)
+    end
+
+    if not description or description == "" then
+        err("missing description when adding a callback to '%s'", name)
+    end
+
+    local prev_descriptions = callback_description[name] or {}
+    for _, desc in ipairs(prev_descriptions) do
+        if desc == description then
+            err("for callback '%s' there already is '%s' added", name, description)
+        end
+    end
+
+    if callback_types[name] == "exclusive" and #prev_descriptions == 1 then
+        err("can't register '%s' as callback '%s' - exclusive callback occupied by '%s'",
+            description, name, prev_descriptions[1])
+    end
+
+    if type(fn) ~= "function" then
+        err("expected Lua function to be added as '%s' for callback '%s'", description, name)
     end
 
     -- add function to callback list for this callback
@@ -251,6 +345,10 @@ function callback.remove_from_callback(name, description)
         end
     end
 
+    if not index then
+        err("can't remove '%s' from callback '%s': not found", description, name)
+    end
+
     table.remove(descriptions, index)
     local fn = table.remove(callback_functions[name], index)
 
@@ -293,9 +391,9 @@ function callback.call_callback(name, ...)
     local functions = callback_functions[name] or {default_functions[name]}
 
     if cbtype == nil then
-        err("cannot call callback '"..name.."' - no such callback exists")
+        err("cannot call callback '%s' - no such callback exists", name)
     elseif cbtype == "exclusive" then
-        -- only one function, atleast default function is guaranteed by
+        -- only one function, at least default function is guaranteed by
         -- create_callback
         return functions[1](...)
     elseif cbtype == "simple" then
@@ -325,23 +423,21 @@ function callback.call_callback(name, ...)
 
     local head = (...)
     local new_head
-    local changed = false
     for _, fn in iter(functions) do
         new_head = fn(head, select(2, ...))
         if new_head == false then
             return false
         elseif new_head ~= true then
             head = new_head
-            changed = true
         end
     end
-    return not changed or head
+    return head
 end
 call_callback = callback.call_callback
 --
 -- Create \"virtual" callbacks `pre/post_mlist_to_hlist_filter` by setting
 -- `mlist_to_hlist` callback. The default behaviour of `mlist_to_hlist` is kept by
--- using a default function, but it can still be overriden by using
+-- using a default function, but it can still be overridden by using
 -- `add_to_callback`.
 default_functions["mlist_to_hlist"] = node.mlist_to_hlist
 callback.create_callback("pre_mlist_to_hlist_filter", "list")
@@ -352,7 +448,7 @@ callback_register("mlist_to_hlist", function(head, ...)
     if new_head == false then
         node.flush_list(head)
         return nil
-    elseif new_head ~= true then
+    else
         head = new_head
     end
     -- mlist_to_hlist means either added functions or standard luatex behavior
@@ -363,26 +459,21 @@ callback_register("mlist_to_hlist", function(head, ...)
     if new_head == false then
         node.flush_list(head)
         return nil
-    elseif new_head ~= true then
-        head = new_head
     end
-    return head
+    return new_head
 end)
 --
 -- For preprocessing boxes just before shipout we define custom callback. This
 -- is used for coloring based on attributes.
 -- There is however a challenge - how to call this callback? We could redefine
 -- `\shipout` and `\pdfxform` (which both run `ship_out` procedure internally),
--- but they would lose their primtive meaning – i.e. `\immediate` wouldn't work
+-- but they would lose their primitive meaning – i.e. `\immediate` wouldn't work
 -- with `\pdfxform`. The compromise is to require anyone to run
 -- \`\_preshipout``<destination box number><box specification>` just before
 -- `\shipout` or `\pdfxform` if they want to call `pre_shipout_filter` (and
 -- achieve colors and possibly more).
 callback.create_callback("pre_shipout_filter", "list")
 
-local tex_setbox = tex.setbox
-local token_scanint = token.scan_int
-local token_scanlist = token.scan_list
 define_lua_command("_preshipout", function()
     local boxnum = token_scanint()
     local head = token_scanlist()
@@ -417,27 +508,76 @@ callback.add_to_callback("input_level_string", function(n)
         return ""
     end
 end, "_tracingmacros")
--- \medskip\secc[lua-pdf-resources] Management of PDF page resources^^M
 --
--- Traditionally, pdf\TeX/ allowed managing PDF page resources (graphics
--- states, patterns, shadings, etc.) using a single toks register,
--- `\pdfpageresources`. This is insufficient due to the expected PDF object
--- structer and also because many \"packages" want to add page resources and
--- thus fight for the access to that register. We add a finer alternative,
--- which allows adding different kinds of resources to a global page resources
--- dictionary. Note that some resource types (fonts and XObjects) are already
--- managed by \LuaTeX/ and shouldn't be added!
+-- \medskip\secc[lua-pdf-utils] PDF object utilities^^M
 --
--- XObject forms can also use resources, but there are several ways to make
--- \LuaTeX/ reference resources from forms. It is hence left up to the user to
--- insert page resources managed by us, if they need them. For that, use
--- `pdf.get_page_resources()`, or the below \TeX/ alternative for that.
+-- The PDF format defines various kinds of \"objects": numbers, names, arrays,
+-- dictionaries. A PDF document is mostly a tree of these objects (i.e. objects
+-- contain other objects), with either direct (\"in-place") or indirect
+-- references.
+--
+-- These objects are saved in the PDF file in a serialized form, often
+-- compressed. While \LuaTeX/ takes care of the compression, and most of the
+-- structure of the PDF format, sometimes we want to insert our own PDF objects
+-- - and there are places where \LuaTeX/ allows us to use our own objects to
+-- insert anything, but it already expects a serialized form of the objects. The
+-- serialized form for some kinds of objects has various formatting and escaping
+-- rules, and is inconvenient to produce in \TeX/ especially when some of the
+-- e.g. names and strings are {\it user input}.
+--
+-- Here we define a couple of utilities for both Lua and \TeX/ that aid with
+-- representation and serialization of objects. The functions are exported, but
+-- may still evolve in incompatible ways.
+--
+-- A reference for the serialization is the \"Syntax" section of the PDF
+-- specification\fnote{\url{https://opensource.adobe.com/dc-acrobat-sdk-docs/pdfstandards/PDF32000_2008.pdf}}.
+--
+-- \`\pdfname``{<name>}` serializes a \"PDF name" to bytes.
+local function name_escape(char)
+    return fmt("#%02X", char:byte())
+end
+--
+local function pdf_name(name)
+    -- Delimiters (i.e. "()<>[]{}/%") and control characters (0x00-0x1f, and 0x7f),
+    -- space (" ", 0x20) as well as number sign ("#", 0x23) characters have to be escaped
+    return "/" .. name:gsub("[%(%)<>%[%]{}/%%%c #]", name_escape)
+end
+optex.pdf_name = pdf_name
+define_lua_command("_pdfname", function()
+    tex_print(-2, pdf_name(token.scan_string()))
+end)
+--
+-- \`\pdfstring``{<string>}` serializes a PDF string to bytes.
+local function pdf_string(str)
+    local out = {"<FEFF"}
+    for _, c in utf8.codes(str) do
+        if c < 0x10000 then
+            out[#out+1] = fmt("%04X", c)
+        else
+            c = c - 0x10000
+            local high = bit32.rshift(c, 10) + 0xD800
+            local low = bit32.band(c, 0x3FF) + 0xDC00
+            out[#out+1] = fmt("%04X%04X", high, low)
+        end
+    end
+    out[#out+1] = ">"
+    return table.concat(out)
+end
+optex.pdf_string = pdf_string
+define_lua_command("_pdfstring", function()
+    tex_print(-2, pdf_string(token.scan_string()))
+end)
+--
+local function pdf_ref(id)
+    return fmt("%d 0 R", id)
+end
+optex.pdf_ref = pdf_ref
 --
 local pdfdict_mt = {
     __tostring = function(dict)
         local out = {"<<"}
         for k, v in pairs(dict) do
-            out[#out+1] = fmt("/%s %s", tostring(k), tostring(v))
+            out[#out+1] = fmt("%s %s", pdf_name(k), tostring(v))
         end
         out[#out+1] = ">>"
         return table.concat(out, "\n")
@@ -448,13 +588,32 @@ local function pdf_dict(t)
 end
 optex.pdf_dict = pdf_dict
 --
+tex.print ("\\_public \\pdfname \\pdfstring ;")
+--
+-- \medskip\secc[lua-pdf-resources] Management of PDF page resources^^M
+--
+-- Traditionally, pdf\TeX/ allowed managing PDF page resources (graphics
+-- states, patterns, shadings, etc.) using a single toks register,
+-- `\pdfpageresources`. This is insufficient due to the expected PDF object
+-- structure and also because many \"packages" want to add page resources and
+-- thus fight for the access to that register. We add a finer alternative,
+-- which allows adding different kinds of resources to a global page resources
+-- dictionary. Note that some resource types (fonts and XObjects) are already
+-- managed by \LuaTeX/ and shouldn't be added!
+--
+-- XObject forms can also use resources, but there are several ways to make
+-- \LuaTeX/ reference resources from forms. It is hence left up to the user to
+-- insert page resources managed by us, if they need them. For that, use
+-- `pdf.get_page_resources()`, or the below \TeX/ alternative for that.
+--
+--
 local resource_dict_objects = {}
 local page_resources = {}
 function pdf.add_page_resource(type, name, value)
     local resources = page_resources[type]
     if not resources then
         local obj = pdf.reserveobj()
-        pdf.setpageresources(fmt("%s /%s %d 0 R", pdf.get_page_resources(), type, obj))
+        pdf.setpageresources(fmt("%s %s %s", pdf.get_page_resources(), pdf_name(type), pdf_ref(obj)))
         resource_dict_objects[type] = obj
         resources = pdf_dict()
         page_resources[type] = resources
@@ -466,14 +625,14 @@ function pdf.get_page_resources()
 end
 --
 -- New \"pseudo" primitives are introduced.
--- \`\_addpageresource``{<type>}{<PDF name>}{<PDF dict>}` adds more reources
+-- \`\_addpageresource``{<type>}{<PDF name>}{<PDF dict>}` adds more resources
 -- of given resource <type> to our data structure.
 -- \`\_pageresources` expands to the saved <type>s and object numbers.
 define_lua_command("_addpageresource", function()
     pdf.add_page_resource(token.scan_string(), token.scan_string(), token.scan_string())
 end)
 define_lua_command("_pageresources", function()
-    tex.print(pdf.get_page_resources())
+    tex_print(pdf.get_page_resources())
 end)
 --
 -- We write the objects with resources to the PDF file in the `finish_pdffile`
@@ -483,7 +642,7 @@ callback.add_to_callback("finish_pdffile", function()
         local obj = resource_dict_objects[type]
         pdf.immediateobj(obj, tostring(dict))
     end
-end)
+end, "_pageresources")
 --
 -- \medskip\secc[lua-colors] Handling of colors and transparency using attributes^^M
 --
@@ -491,36 +650,6 @@ end)
 -- to them. We do this by intercepting \TeX/ just before it ships out a page and
 -- inject PDF literals according to attributes.
 --
-local node_id = node.id
-local node_subtype = node.subtype
-local glyph_id = node_id("glyph")
-local rule_id = node_id("rule")
-local glue_id = node_id("glue")
-local hlist_id = node_id("hlist")
-local vlist_id = node_id("vlist")
-local disc_id = node_id("disc")
-local whatsit_id = node_id("whatsit")
-local pdfliteral_id = node_subtype("pdf_literal")
-local pdfsave_id = node_subtype("pdf_save")
-local pdfrestore_id = node_subtype("pdf_restore")
-local token_getmacro = token.get_macro
-
-local direct = node.direct
-local todirect = direct.todirect
-local tonode = direct.tonode
-local getfield = direct.getfield
-local setfield = direct.setfield
-local getwhd = direct.getwhd
-local getid = direct.getid
-local getlist = direct.getlist
-local setlist = direct.setlist
-local getleader = direct.getleader
-local getattribute = direct.get_attribute
-local insertbefore = direct.insert_before
-local copy = direct.copy
-local traverse = direct.traverse
-local one_bp = tex.sp("1bp")
---
 -- The attribute for coloring is allocated in `colors.opm`
 local color_attribute = registernumber("_colorattr")
 local transp_attribute = registernumber("_transpattr")
@@ -540,7 +669,7 @@ optex.directpdfliteral = pdfliteral
 -- The function {\Red`colorize`}`(head, current, current_stroke, current_tr)`
 -- goes through a node list and injects PDF literals according to attributes.
 -- Its arguments are the head of the list to be colored and the current color
--- for fills and strokes and the current trasparency attribute.
+-- for fills and strokes and the current transparency attribute.
 -- It is a recursive function – nested
 -- horizontal and vertical lists are handled in the same way. Only the
 -- attributes of “content” nodes (glyphs, rules, etc.) matter. Users drawing
@@ -575,7 +704,9 @@ optex.directpdfliteral = pdfliteral
 -- \TeX/ works with nodes.
 local function is_color_needed(head, n, id, subtype) -- returns fill, stroke color needed
     if id == glyph_id then
-        return true, false
+        local font_id = getfont(n)
+        local mode = font_getfont(font_id).mode
+        return true, mode == 1 or mode == 2
     elseif id == glue_id then
         n = getleader(n)
         if n then
@@ -664,11 +795,7 @@ end, "_colors")
 -- attribute. On top of that, we take care of transparency resources ourselves.
 --
 -- The hook has to be registered {\em after} `luaotfload` is loaded.
-local setattribute = direct.set_attribute
-local token_setmacro = token.set_macro
 local color_count = registernumber("_colorcnt")
-local tex_getcount, tex_setcount = tex.getcount, tex.setcount
---
 local function set_node_color(n, color) -- "1 0 0 rg" or "0 g", etc.
     local attr = tonumber(token_getmacro("_color::"..color))
     if not attr then
@@ -716,7 +843,18 @@ function optex.hook_into_luaotfload()
     end)
 end
 
+-- \`\_beglocalcontrol` `<tokens>` `\_endlocalcontrol` runs `<tokens>` fully at
+-- expand processor level despite the fact that `<tokens>` processes unexpandable commands.
+
+define_lua_command("_beglocalcontrol", function()
+	return tex.runtoks(token.get_next)
+end)
+
    -- History:
+   -- 2024-12-18 \pdfstring etc. introduced
+   -- 2024-09-06 raw_ht() implemented
+   -- 2024-06-02 more checking in add_to_callback and remove_from_callback
+   -- 2024-02-18 \_beglocalcontrol added
    -- 2022-08-25 expose some useful functions in `optex` namespace
    -- 2022-08-24 luaotfload transparency with attributes added
    -- 2022-03-07 transparency in the colorize() function, current_tr added
